%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  Purpose
%
%       Driver for example in Exercise 4.3. 
%
%       See Exercise 4.3 in J. Nocedal and S. J. Wright,
%       "Numerical Optimization", Springer-Verlag, 1999
%
%  Functions called: func.m, gradient.m, hessian.m
%
%
%  Written by L. N. Vicente.
%  April 14, 2000.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Begin.
%
% Clear work space and screen.
  clf;
  clear;
%
% Set formats.
  format compact;
  format long;
%
% Initializations.
%
% fid   file identifier for output file. No output if fid < 0.
%       Use fid = 1 for standard output (usually screen).
  fid = 1; 
%
% Parameters for line search.
  c1  = 10^(-4);
  rho = 0.5;
%
% Initial points.
%
  n = 10;
%
  for i = 1:n
      x0(2*i-1) = 0;
      x0(2*i)   = 0.5;
  end
  x0 = x0';
%
% Newton's method with a backtracking line search.
  fl = flops;
  iter  = 0; ifunc=0; ig=0; iH=0;
  x     = x0;
  [f]   = func(x);
  ifunc = ifunc+1;
  [g]   = gradient(x);
  ig    = ig+1;
  while ( norm(g) > sqrt(eps) )
          [H]       = hessian(x);
          iH        = iH+1;
          p         = - H\g;
          alpha     = 1;
          xalphap   = x + alpha*p;
          [falphap] = func(xalphap);
          ifunc     = ifunc+1;
          linearf   = f+c1*alpha*g'*p;
          while ( falphap > linearf )
%                 Backtrack.
                  alpha     = rho*alpha; 
                  xalphap   = x + alpha*p;
                  [falphap] = func(xalphap); 
                  ifunc     = ifunc+1;
                  linearf   = f+c1*alpha*g'*p;
          end
          x    = xalphap;
          f    = falphap;
          [g]  = gradient(x);
          ig   = ig+1;
          iter = iter + 1;
  end
%
  fle = flops - fl;
  normg = norm(g);
  condH = cond(H);
% Printing.
  fprintf( fid,'\nNewton''s method with a backtracking line search \n');
  fprintf( fid,'\n|  iter  | fevals | gevals | Hevals |   norm(g)  |   cond(H)  |    flops   | \n');
  format_str = '| %6d | %6d | %6d | %6d | %9.4e | %9.4e | %10d | \n \n';
  fprintf( fid, format_str, iter, ifunc, ig, iH, normg, condH, fle);
%
% BFGS method with a backtracking line search.
  fl = flops;
  iter = 0; ifunc=0; ig=0; iINVBFGS=0;
  x       = x0;
  [f]     = func(x);
  ifunc   = ifunc+1;
  [g]     = gradient(x);
  ig      = ig+1;
  INVBFGS = eye(2*n,2*n);
  while ( norm(g) > sqrt(eps) )
          p         = - INVBFGS * g;
          alpha     = 1;
          xalphap   = x + alpha*p;
          [falphap] = func(xalphap);
          ifunc     = ifunc+1;
          linearf   = f+c1*alpha*g'*p;
          while ( falphap > linearf )
%                 Backtrack.
                  alpha     = rho*alpha;
                  xalphap   = x + alpha*p;
                  [falphap] = func(xalphap);
                  ifunc     = ifunc+1;
                  linearf   = f+c1*alpha*g'*p;
          end
          xnew   = xalphap;
          f      = falphap;
          [gnew] = gradient(xnew);
          ig     = ig+1;
%
          y = gnew - g;
          s = xnew - x;
%
%         BFGS update.
          rhoBFGS  = 1/(y'*s);
          INVBFGS  = INVBFGS - rhoBFGS * ( (INVBFGS * y) * s' + s * ( y' * INVBFGS ) ) ...
                     + (rhoBFGS + rhoBFGS^2 * (y' * INVBFGS * y) ) * ( s * s' );
          iINVBFGS = iINVBFGS+1;
%
          if ( iter == 0 ) INVBFGS = ((y'*s)/(y'*y)) * eye(2*n,2*n); end
%
          x    =  xnew;
          g    =  gnew;
          iter = iter + 1;
  end
  fle = flops - fl;
  normg = norm(g);
  condH = cond(INVBFGS);
% Printing.
  fprintf( fid,'BFGS method with a backtracking line search \n');
  fprintf( fid,'\n|  iter  | fevals | gevals |  BFGSs |   norm(g)  |  cn(IBFGS) |    flops   | \n');
  format_str = '| %6d | %6d | %6d | %6d | %9.4e | %9.4e | %10d | \n \n';
  fprintf( fid, format_str, iter, ifunc, ig, iINVBFGS, normg, condH, fle);
%
% Newton's method.
  fl = flops;
  iter = 0; ifunc=0; ig=0; iH=0;
  x    = x0;    
  [g]  = gradient(x);
  ig   = ig+1;
  while ( norm(g) > sqrt(eps) )
          [H]  = hessian(x);
          iH   = iH+1;
          x    = x - H\g;
          [g]  = gradient(x);
          ig   = ig+1;
          iter = iter + 1;
  end
  fle = flops - fl;
  normg = norm(g);
  condH = cond(H);
% Printing.
  fprintf( fid,'Newton''s method \n');
  fprintf( fid,'\n|  iter  | fevals | gevals | Hevals |   norm(g)  |   cond(H)  |    flops   | \n');
  format_str = '| %6d | %6d | %6d | %6d | %9.4e | %9.4e | %10d | \n \n';
  fprintf( fid, format_str, iter, ifunc, ig, iH, normg, condH, fle);
%
% BFGS method.
  fl = flops;
  iter = 0; ifunc=0; ig=0; iINVBFGS=0; 
  x       = x0;
  [g]     = gradient(x);
  ig      = ig+1;
  INVBFGS = eye(2*n,2*n);
  while ( norm(g) > sqrt(eps) )
          xnew   = x - INVBFGS * g;
          [gnew] = gradient(xnew);
          ig     = ig+1;
%          
          y = gnew - g;
          s = xnew - x; 
%           
%         BFGS update.
          rhoBFGS  = 1/(y'*s);
          v        = INVBFGS * y;
          INVBFGS  = INVBFGS - rhoBFGS * ( v * s' + s * v' ) ...
                     + (rhoBFGS + rhoBFGS^2 * (y' * v) ) * ( s * s' );
          iINVBFGS = iINVBFGS+1;
%
          if ( iter == 0 ) INVBFGS = ((y'*s)/(y'*y)) * eye(2*n,2*n); end
%
          x    =  xnew;
          g    =  gnew;
          iter = iter + 1;
  end
  fle = flops - fl;
  normg = norm(g);
  condH = cond(INVBFGS);
% Printing.
  fprintf( fid,'BFGS method \n');
  fprintf( fid,'\n|  iter  | fevals | gevals |  BFGSs |   norm(g)  |  cn(IBFGS) |    flops   | \n');
  format_str = '| %6d | %6d | %6d | %6d | %9.4e | %9.4e | %10d | \n \n';
  fprintf( fid, format_str, iter, ifunc, ig, iINVBFGS, normg, condH, fle);
%
% End of driver for the minimization of the function in Exercise 4.3 of Nocedal and Wright.
